# Root of the SAM package where we expose public classes & methods for other consumers of this SAM Translator to use.
# This is essentially our Public API
#

__all__ = ["Translator", "ManagedPolicyLoader"]

from samtranslator.translator.managed_policy_translator import ManagedPolicyLoader
from samtranslator.translator.translator import Translator



print("import: 'samtranslator'")
import samtranslator

print("import: 'samtranslator.intrinsics'")
import samtranslator.intrinsics

print("import: 'samtranslator.model'")
import samtranslator.model

print("import: 'samtranslator.parser'")
import samtranslator.parser

print("import: 'samtranslator.plugins'")
import samtranslator.plugins

print("import: 'samtranslator.policy_template_processor'")
import samtranslator.policy_template_processor

print("import: 'samtranslator.public'")
import samtranslator.public

print("import: 'samtranslator.sdk'")
import samtranslator.sdk

print("import: 'samtranslator.swagger'")
import samtranslator.swagger

print("import: 'samtranslator.translator'")
import samtranslator.translator

print("import: 'samtranslator.validator'")
import samtranslator.validator




import json
import time
from re import match
from typing import Any, Dict, List, Optional, Union

from samtranslator.model import GeneratedProperty, Resource
from samtranslator.model.exceptions import InvalidResourceException
from samtranslator.model.intrinsics import fnSub, ref
from samtranslator.model.types import PassThrough
from samtranslator.translator import logical_id_generator
from samtranslator.translator.arn_generator import ArnGenerator
from samtranslator.utils.py27hash_fix import Py27Dict, Py27UniStr
from samtranslator.validator.value_validator import sam_expect


class ApiGatewayRestApi(Resource):
    resource_type = "AWS::ApiGateway::RestApi"
    property_types = {
        "Body": GeneratedProperty(),
        "BodyS3Location": GeneratedProperty(),
        "CloneFrom": GeneratedProperty(),
        "Description": GeneratedProperty(),
        "FailOnWarnings": GeneratedProperty(),
        "Name": GeneratedProperty(),
        "Parameters": GeneratedProperty(),
        "EndpointConfiguration": GeneratedProperty(),
        "BinaryMediaTypes": GeneratedProperty(),
        "MinimumCompressionSize": GeneratedProperty(),
        "Mode": GeneratedProperty(),
        "ApiKeySourceType": GeneratedProperty(),
        "Tags": GeneratedProperty(),
    }

    Body: Optional[Dict[str, Any]]
    BodyS3Location: Optional[Dict[str, Any]]
    CloneFrom: Optional[PassThrough]
    Description: Optional[PassThrough]
    FailOnWarnings: Optional[PassThrough]
    Name: Optional[PassThrough]
    Parameters: Optional[Dict[str, Any]]
    EndpointConfiguration: Optional[Dict[str, Any]]
    BinaryMediaTypes: Optional[List[Any]]
    MinimumCompressionSize: Optional[PassThrough]
    Mode: Optional[PassThrough]
    ApiKeySourceType: Optional[PassThrough]
    Tags: Optional[PassThrough]

    runtime_attrs = {"rest_api_id": lambda self: ref(self.logical_id)}


class ApiGatewayStage(Resource):
    resource_type = "AWS::ApiGateway::Stage"
    property_types = {
        "AccessLogSetting": GeneratedProperty(),
        "CacheClusterEnabled": GeneratedProperty(),
        "CacheClusterSize": GeneratedProperty(),
        "CanarySetting": GeneratedProperty(),
        "ClientCertificateId": GeneratedProperty(),
        "DeploymentId": GeneratedProperty(),
        "Description": GeneratedProperty(),
        "RestApiId": GeneratedProperty(),
        "StageName": GeneratedProperty(),
        "Tags": GeneratedProperty(),
        "TracingEnabled": GeneratedProperty(),
        "Variables": GeneratedProperty(),
        "MethodSettings": GeneratedProperty(),
    }

    runtime_attrs = {"stage_name": lambda self: ref(self.logical_id)}

    def update_deployment_ref(self, deployment_logical_id: str) -> None:
        self.DeploymentId = ref(deployment_logical_id)


class ApiGatewayAccount(Resource):
    resource_type = "AWS::ApiGateway::Account"
    property_types = {
        "CloudWatchRoleArn": GeneratedProperty(),
    }


class ApiGatewayDeployment(Resource):
    _X_HASH_DELIMITER = "||"

    resource_type = "AWS::ApiGateway::Deployment"
    property_types = {
        "Description": GeneratedProperty(),
        "RestApiId": GeneratedProperty(),
        "StageDescription": GeneratedProperty(),
        "StageName": GeneratedProperty(),
    }

    runtime_attrs = {"deployment_id": lambda self: ref(self.logical_id)}

    def make_auto_deployable(  # noqa: PLR0913
        self,
        stage: ApiGatewayStage,
        openapi_version: Optional[Union[Dict[str, Any], str]] = None,
        swagger: Optional[Dict[str, Any]] = None,
        domain: Optional[Dict[str, Any]] = None,
        redeploy_restapi_parameters: Optional[Any] = None,
        always_deploy: Optional[bool] = False,
    ) -> None:
        """
        Sets up the resource such that it will trigger a re-deployment when Swagger changes or always_deploy is true
        or the openapi version changes or a domain resource changes.

        :param stage: The ApiGatewayStage object which will be re-deployed
        :param swagger: Dictionary containing the Swagger definition of the API
        :param openapi_version: string containing value of OpenApiVersion flag in the template
        :param domain: Dictionary containing the custom domain configuration for the API
        :param redeploy_restapi_parameters: Dictionary containing the properties for which rest api will be redeployed
        """
        if not swagger:
            return

        # CloudFormation does NOT redeploy the API unless it has a new deployment resource
        # that points to latest RestApi resource. Append a hash of Swagger Body location to
        # redeploy only when the API data changes. First 10 characters of hash is good enough
        # to prevent redeployment when API has not changed

        # NOTE: `str(swagger)` is for backwards compatibility. Changing it to a JSON or something will break compat
        hash_input = [str(swagger)]
        if openapi_version:
            hash_input.append(str(openapi_version))
        if domain:
            hash_input.append(json.dumps(domain))
        function_names = redeploy_restapi_parameters.get("function_names") if redeploy_restapi_parameters else None
        # The deployment logical id is <api logicalId> + "Deployment"
        # The keyword "Deployment" is removed and all the function names associated with api is obtained
        if function_names and function_names.get(self.logical_id[:-10], None):
            hash_input.append(function_names.get(self.logical_id[:-10], ""))
        if always_deploy:
            # We just care that the hash changes every time
            # Using int so tests are a little more robust; don't think the Python spec defines default precision
            hash_input = [str(int(time.time()))]
        data = self._X_HASH_DELIMITER.join(hash_input)
        generator = logical_id_generator.LogicalIdGenerator(self.logical_id, data)
        self.logical_id = generator.gen()
        digest = generator.get_hash(length=40)
        self.Description = f"RestApi deployment id: {digest}"
        stage.update_deployment_ref(self.logical_id)


class ApiGatewayResponse:
    ResponseParameterProperties = ["Headers", "Paths", "QueryStrings"]

    def __init__(
        self,
        api_logical_id: str,
        response_parameters: Optional[Dict[str, Any]] = None,
        response_templates: Optional[PassThrough] = None,
        status_code: Optional[str] = None,
    ) -> None:
        if response_parameters:
            # response_parameters has been validated in ApiGenerator._add_gateway_responses()
            for response_parameter_key in response_parameters:
                if response_parameter_key not in ApiGatewayResponse.ResponseParameterProperties:
                    raise InvalidResourceException(
                        api_logical_id, f"Invalid gateway response parameter '{response_parameter_key}'"
                    )

        status_code_str = self._status_code_string(status_code)  # type: ignore[no-untyped-call]
        # status_code must look like a status code, if present. Let's not be judgmental; just check 0-999.
        if status_code and not match(r"^[0-9]{1,3}$", status_code_str):
            raise InvalidResourceException(api_logical_id, "Property 'StatusCode' must be numeric")

        self.api_logical_id = api_logical_id
        # Defaults to Py27Dict() as these will go into swagger
        self.response_parameters = response_parameters or Py27Dict()
        self.response_templates = response_templates or Py27Dict()
        self.status_code = status_code_str

    def generate_swagger(self) -> Py27Dict:
        # Applying Py27Dict here as this goes into swagger
        swagger = Py27Dict()
        swagger["responseParameters"] = self._add_prefixes(self.response_parameters)
        swagger["responseTemplates"] = self.response_templates

        # Prevent "null" being written.
        if self.status_code:
            swagger["statusCode"] = self.status_code

        return swagger

    def _add_prefixes(self, response_parameters: Dict[str, Any]) -> Dict[str, str]:
        GATEWAY_RESPONSE_PREFIX = "gatewayresponse."
        # applying Py27Dict as this is part of swagger
        prefixed_parameters = Py27Dict()

        parameter_prefix_pairs = [("Headers", "header."), ("Paths", "path."), ("QueryStrings", "querystring.")]
        for parameter_property_name, prefix in parameter_prefix_pairs:
            parameter_property_value = response_parameters.get(parameter_property_name, {})
            sam_expect(
                parameter_property_value, self.api_logical_id, f"ResponseParameters.{parameter_property_name}"
            ).to_be_a_map()
            for key, value in parameter_property_value.items():
                param_key = GATEWAY_RESPONSE_PREFIX + prefix + key
                if isinstance(key, Py27UniStr):
                    # if key is from template, we need to convert param_key to Py27UniStr
                    param_key = Py27UniStr(param_key)
                prefixed_parameters[param_key] = value

        return prefixed_parameters

    def _status_code_string(self, status_code):  # type: ignore[no-untyped-def]
        return None if status_code is None else str(status_code)


class ApiGatewayDomainName(Resource):
    resource_type = "AWS::ApiGateway::DomainName"
    property_types = {
        "RegionalCertificateArn": GeneratedProperty(),
        "DomainName": GeneratedProperty(),
        "EndpointConfiguration": GeneratedProperty(),
        "MutualTlsAuthentication": GeneratedProperty(),
        "SecurityPolicy": GeneratedProperty(),
        "CertificateArn": GeneratedProperty(),
        "Tags": GeneratedProperty(),
        "OwnershipVerificationCertificateArn": GeneratedProperty(),
    }

    RegionalCertificateArn: Optional[PassThrough]
    DomainName: PassThrough
    EndpointConfiguration: Optional[PassThrough]
    MutualTlsAuthentication: Optional[Dict[str, Any]]
    SecurityPolicy: Optional[PassThrough]
    CertificateArn: Optional[PassThrough]
    Tags: Optional[PassThrough]
    OwnershipVerificationCertificateArn: Optional[PassThrough]


class ApiGatewayBasePathMapping(Resource):
    resource_type = "AWS::ApiGateway::BasePathMapping"
    property_types = {
        "BasePath": GeneratedProperty(),
        "DomainName": GeneratedProperty(),
        "RestApiId": GeneratedProperty(),
        "Stage": GeneratedProperty(),
    }


class ApiGatewayUsagePlan(Resource):
    resource_type = "AWS::ApiGateway::UsagePlan"
    property_types = {
        "ApiStages": GeneratedProperty(),
        "Description": GeneratedProperty(),
        "Quota": GeneratedProperty(),
        "Tags": GeneratedProperty(),
        "Throttle": GeneratedProperty(),
        "UsagePlanName": GeneratedProperty(),
    }
    runtime_attrs = {"usage_plan_id": lambda self: ref(self.logical_id)}


class ApiGatewayUsagePlanKey(Resource):
    resource_type = "AWS::ApiGateway::UsagePlanKey"
    property_types = {
        "KeyId": GeneratedProperty(),
        "KeyType": GeneratedProperty(),
        "UsagePlanId": GeneratedProperty(),
    }


class ApiGatewayApiKey(Resource):
    resource_type = "AWS::ApiGateway::ApiKey"
    property_types = {
        "CustomerId": GeneratedProperty(),
        "Description": GeneratedProperty(),
        "Enabled": GeneratedProperty(),
        "GenerateDistinctId": GeneratedProperty(),
        "Name": GeneratedProperty(),
        "Tags": GeneratedProperty(),
        "StageKeys": GeneratedProperty(),
        "Value": GeneratedProperty(),
    }

    runtime_attrs = {"api_key_id": lambda self: ref(self.logical_id)}


class ApiGatewayAuthorizer:
    _VALID_FUNCTION_PAYLOAD_TYPES = [None, "TOKEN", "REQUEST"]

    def __init__(  # type: ignore[no-untyped-def]# noqa: PLR0913
        self,
        api_logical_id=None,
        name=None,
        user_pool_arn=None,
        function_arn=None,
        identity=None,
        function_payload_type: Optional[str] = None,
        function_invoke_role=None,
        is_aws_iam_authorizer=False,
        authorization_scopes=None,
        disable_function_default_permissions=False,
    ):
        if authorization_scopes is None:
            authorization_scopes = []

        self.api_logical_id = api_logical_id
        self.name = name
        self.user_pool_arn = user_pool_arn
        self.function_arn = function_arn
        self.identity = identity
        self.function_payload_type = function_payload_type
        self.function_invoke_role = function_invoke_role
        self.is_aws_iam_authorizer = is_aws_iam_authorizer
        self.authorization_scopes = authorization_scopes
        self.disable_function_default_permissions = disable_function_default_permissions

        if function_payload_type not in ApiGatewayAuthorizer._VALID_FUNCTION_PAYLOAD_TYPES:
            raise InvalidResourceException(
                api_logical_id,
                f"{name} Authorizer has invalid 'FunctionPayloadType': {function_payload_type}.",
            )

        if function_payload_type == "REQUEST" and self._is_missing_identity_source(identity):
            raise InvalidResourceException(
                api_logical_id,
                f"{name} Authorizer must specify Identity with at least one "
                "of Headers, QueryStrings, StageVariables, or Context.",
            )

        if authorization_scopes is not None:
            sam_expect(authorization_scopes, api_logical_id, f"Authorizers.{name}.AuthorizationScopes").to_be_a_list()

        if disable_function_default_permissions is not None:
            sam_expect(
                disable_function_default_permissions,
                api_logical_id,
                f"Authorizers.{name}.DisableFunctionDefaultPermissions",
            ).to_be_a_bool()

    def _is_missing_identity_source(self, identity: Dict[str, Any]) -> bool:
        if not identity:
            return True

        sam_expect(identity, self.api_logical_id, f"Authorizer.{self.name}.Identity").to_be_a_map()

        headers = identity.get("Headers")
        query_strings = identity.get("QueryStrings")
        stage_variables = identity.get("StageVariables")
        context = identity.get("Context")
        ttl = identity.get("ReauthorizeEvery")

        required_properties_missing = not headers and not query_strings and not stage_variables and not context

        if ttl is None:
            return required_properties_missing
        try:
            ttl_int = int(ttl)
        # this will catch if and not convertable to an int
        except (TypeError, ValueError):
            # previous behavior before trying to read ttl
            return required_properties_missing

        # If we can resolve ttl, attempt to see if things are valid
        return ttl_int > 0 and required_properties_missing

    def generate_swagger(self) -> Py27Dict:
        authorizer_type = self._get_type()
        APIGATEWAY_AUTHORIZER_KEY = "x-amazon-apigateway-authorizer"
        swagger = Py27Dict()
        swagger["type"] = "apiKey"
        swagger["name"] = self._get_swagger_header_name()
        swagger["in"] = "header"
        swagger["x-amazon-apigateway-authtype"] = self._get_swagger_authtype()

        if authorizer_type == "COGNITO_USER_POOLS":
            authorizer_dict = Py27Dict()
            authorizer_dict["type"] = self._get_swagger_authorizer_type()
            authorizer_dict["providerARNs"] = self._get_user_pool_arn_array()
            swagger[APIGATEWAY_AUTHORIZER_KEY] = authorizer_dict

        elif authorizer_type == "LAMBDA":
            swagger[APIGATEWAY_AUTHORIZER_KEY] = Py27Dict({"type": self._get_swagger_authorizer_type()})
            partition = ArnGenerator.get_partition_name()
            resource = "lambda:path/2015-03-31/functions/${__FunctionArn__}/invocations"
            authorizer_uri = fnSub(
                ArnGenerator.generate_arn(
                    partition=partition, service="apigateway", resource=resource, include_account_id=False
                ),
                {"__FunctionArn__": self.function_arn},
            )

            swagger[APIGATEWAY_AUTHORIZER_KEY]["authorizerUri"] = authorizer_uri
            reauthorize_every = self._get_reauthorize_every()
            function_invoke_role = self._get_function_invoke_role()

            if reauthorize_every is not None:
                swagger[APIGATEWAY_AUTHORIZER_KEY]["authorizerResultTtlInSeconds"] = reauthorize_every

            if function_invoke_role:
                swagger[APIGATEWAY_AUTHORIZER_KEY]["authorizerCredentials"] = function_invoke_role

            if self._get_function_payload_type() == "REQUEST":
                identity_source = self._get_identity_source()
                if identity_source:
                    swagger[APIGATEWAY_AUTHORIZER_KEY]["identitySource"] = self._get_identity_source()

        # Authorizer Validation Expression is only allowed on COGNITO_USER_POOLS and LAMBDA_TOKEN
        is_lambda_token_authorizer = authorizer_type == "LAMBDA" and self._get_function_payload_type() == "TOKEN"

        if authorizer_type == "COGNITO_USER_POOLS" or is_lambda_token_authorizer:
            identity_validation_expression = self._get_identity_validation_expression()

            if identity_validation_expression:
                swagger[APIGATEWAY_AUTHORIZER_KEY]["identityValidationExpression"] = identity_validation_expression

        return swagger

    def _get_identity_validation_expression(self) -> Optional[PassThrough]:
        return self.identity and self.identity.get("ValidationExpression")

    @staticmethod
    def _build_identity_source_item(item_prefix: str, prop_value: str) -> str:
        item = item_prefix + prop_value
        if isinstance(prop_value, Py27UniStr):
            return Py27UniStr(item)
        return item

    def _build_identity_source_item_array(self, prop_key: str, item_prefix: str) -> List[str]:
        arr: List[str] = []
        prop_value_list = self.identity.get(prop_key)
        if prop_value_list:
            prop_path = f"Auth.Authorizers.{self.name}.Identity.{prop_key}"
            sam_expect(prop_value_list, self.api_logical_id, prop_path).to_be_a_list()
            for index, prop_value in enumerate(prop_value_list):
                sam_expect(prop_value, self.api_logical_id, f"{prop_path}[{index}]").to_be_a_string()
                arr.append(self._build_identity_source_item(item_prefix, prop_value))
        return arr

    def _get_identity_source(self) -> str:
        key_prefix_pairs = [
            ("Headers", "method.request.header."),
            ("QueryStrings", "method.request.querystring."),
            ("StageVariables", "stageVariables."),
            ("Context", "context."),
        ]

        identity_source_array = []
        for prop_key, item_prefix in key_prefix_pairs:
            identity_source_array.extend(self._build_identity_source_item_array(prop_key, item_prefix))

        identity_source = ", ".join(identity_source_array)
        if any(isinstance(i, Py27UniStr) for i in identity_source_array):
            # Convert identity_source to Py27UniStr if any part of it is Py27UniStr
            return Py27UniStr(identity_source)

        return identity_source

    def _get_user_pool_arn_array(self) -> List[PassThrough]:
        return self.user_pool_arn if isinstance(self.user_pool_arn, list) else [self.user_pool_arn]

    def _get_swagger_header_name(self) -> Optional[str]:
        authorizer_type = self._get_type()
        payload_type = self._get_function_payload_type()

        if authorizer_type == "LAMBDA" and payload_type == "REQUEST":
            return "Unused"

        return self._get_identity_header()

    def _get_type(self) -> str:
        if self.is_aws_iam_authorizer:
            return "AWS_IAM"

        if self.user_pool_arn:
            return "COGNITO_USER_POOLS"

        return "LAMBDA"

    def _get_identity_header(self) -> Optional[str]:
        if self.identity and not isinstance(self.identity, dict):
            raise InvalidResourceException(
                self.api_logical_id,
                "Auth.Authorizers.<Authorizer>.Identity must be a dict (LambdaTokenAuthorizationIdentity, "
                "LambdaRequestAuthorizationIdentity or CognitoAuthorizationIdentity).",
            )

        if not self.identity or not self.identity.get("Header"):
            return "Authorization"

        return self.identity.get("Header")

    def _get_reauthorize_every(self) -> Optional[PassThrough]:
        if not self.identity:
            return None

        return self.identity.get("ReauthorizeEvery")

    def _get_function_invoke_role(self) -> Optional[PassThrough]:
        if not self.function_invoke_role or self.function_invoke_role == "NONE":
            return None

        return self.function_invoke_role

    def _get_swagger_authtype(self) -> str:
        authorizer_type = self._get_type()
        if authorizer_type == "AWS_IAM":
            return "awsSigv4"

        if authorizer_type == "COGNITO_USER_POOLS":
            return "cognito_user_pools"

        return "custom"

    def _get_function_payload_type(self) -> str:
        return "TOKEN" if not self.function_payload_type else self.function_payload_type

    def _get_swagger_authorizer_type(self) -> Optional[str]:
        authorizer_type = self._get_type()

        if authorizer_type == "COGNITO_USER_POOLS":
            return "cognito_user_pools"

        payload_type = self._get_function_payload_type()

        if payload_type == "REQUEST":
            return "request"

        if payload_type == "TOKEN":
            return "token"

        return None  # should we raise validation error here?



from samtranslator.model import GeneratedProperty, Resource
from samtranslator.model.intrinsics import fnGetAtt, ref


class S3Bucket(Resource):
    resource_type = "AWS::S3::Bucket"
    property_types = {
        "AccessControl": GeneratedProperty(),
        "AccelerateConfiguration": GeneratedProperty(),
        "AnalyticsConfigurations": GeneratedProperty(),
        "BucketEncryption": GeneratedProperty(),
        "BucketName": GeneratedProperty(),
        "CorsConfiguration": GeneratedProperty(),
        "IntelligentTieringConfigurations": GeneratedProperty(),
        "InventoryConfigurations": GeneratedProperty(),
        "LifecycleConfiguration": GeneratedProperty(),
        "LoggingConfiguration": GeneratedProperty(),
        "MetricsConfigurations": GeneratedProperty(),
        "NotificationConfiguration": GeneratedProperty(),
        "ObjectLockConfiguration": GeneratedProperty(),
        "ObjectLockEnabled": GeneratedProperty(),
        "OwnershipControls": GeneratedProperty(),
        "PublicAccessBlockConfiguration": GeneratedProperty(),
        "ReplicationConfiguration": GeneratedProperty(),
        "Tags": GeneratedProperty(),
        "VersioningConfiguration": GeneratedProperty(),
        "WebsiteConfiguration": GeneratedProperty(),
    }

    runtime_attrs = {"name": lambda self: ref(self.logical_id), "arn": lambda self: fnGetAtt(self.logical_id, "Arn")}



from __future__ import annotations

from typing import Any, Dict, Literal, Optional

from samtranslator.internal.schema_source.aws_serverless_connector import EmbeddedConnector
from samtranslator.internal.schema_source.common import (
    BaseModel,
    PassThroughProp,
    ResourceAttributes,
    get_prop,
    passthrough_prop,
)

PROPERTIES_STEM = "sam-resource-simpletable"
PRIMARY_KEY_STEM = "sam-property-simpletable-primarykeyobject"

primarykey = get_prop(PRIMARY_KEY_STEM)
properties = get_prop(PROPERTIES_STEM)


class PrimaryKey(BaseModel):
    Name: PassThroughProp = passthrough_prop(
        PRIMARY_KEY_STEM,
        "Name",
        ["AWS::DynamoDB::Table.AttributeDefinition", "AttributeName"],
    )
    Type: PassThroughProp = passthrough_prop(
        PRIMARY_KEY_STEM,
        "Type",
        ["AWS::DynamoDB::Table.AttributeDefinition", "AttributeType"],
    )


SSESpecification = Optional[PassThroughProp]


class Properties(BaseModel):
    PointInTimeRecoverySpecification: Optional[PassThroughProp]  # TODO: add docs
    PrimaryKey: Optional[PrimaryKey] = properties("PrimaryKey")
    ProvisionedThroughput: Optional[PassThroughProp] = passthrough_prop(
        PROPERTIES_STEM,
        "ProvisionedThroughput",
        ["AWS::DynamoDB::Table", "Properties", "ProvisionedThroughput"],
    )
    SSESpecification: Optional[SSESpecification] = passthrough_prop(
        PROPERTIES_STEM,
        "SSESpecification",
        ["AWS::DynamoDB::Table", "Properties", "SSESpecification"],
    )
    TableName: Optional[PassThroughProp] = passthrough_prop(
        PROPERTIES_STEM,
        "TableName",
        ["AWS::DynamoDB::Table", "Properties", "TableName"],
    )
    Tags: Optional[Dict[str, Any]] = properties("Tags")


class Globals(BaseModel):
    SSESpecification: Optional[SSESpecification] = passthrough_prop(
        PROPERTIES_STEM,
        "SSESpecification",
        ["AWS::DynamoDB::Table", "Properties", "SSESpecification"],
    )


class Resource(ResourceAttributes):
    Type: Literal["AWS::Serverless::SimpleTable"]
    Properties: Optional[Properties]
    Connectors: Optional[Dict[str, EmbeddedConnector]]



from typing import Any, Dict

from samtranslator.metrics.method_decorator import cw_timer
from samtranslator.model.exceptions import InvalidResourceAttributeTypeException
from samtranslator.plugins.globals.globals import Globals, InvalidGlobalsSectionException
from samtranslator.public.exceptions import InvalidDocumentException
from samtranslator.public.plugins import BasePlugin
from samtranslator.public.sdk.template import SamTemplate

_API_RESOURCE = "AWS::Serverless::Api"


class GlobalsPlugin(BasePlugin):
    """
    Plugin to process Globals section of a SAM template before the template is translated to CloudFormation.
    """

    @cw_timer(prefix="Plugin-Globals")
    def on_before_transform_template(self, template_dict: Dict[str, Any]) -> None:
        """
        Hook method that runs before a template gets transformed. In this method, we parse and process Globals section
        from the template (if present).

        :param dict template_dict: SAM template as a dictionary
        """
        try:
            global_section = Globals(template_dict)
        except InvalidGlobalsSectionException as ex:
            raise InvalidDocumentException([ex]) from ex

        # For each resource in template, try and merge with Globals if necessary
        template = SamTemplate(template_dict)
        for logicalId, resource in template.iterate():
            try:
                resource.properties = global_section.merge(
                    str(resource.type), resource.properties, logicalId, resource.ignore_globals
                )
            except InvalidResourceAttributeTypeException as ex:
                raise InvalidDocumentException([ex]) from ex
            template.set(logicalId, resource)

        # Remove the Globals section from template if necessary
        Globals.del_section(template_dict)

        # If there was a global openApiVersion flag, check and convert swagger
        # to the right version
        Globals.fix_openapi_definitions(template_dict)



import copy
from typing import Any, List, Optional, Union, cast


def as_array(x: Any) -> List[Any]:
    """Convert value to list if it already isn't."""
    return x if isinstance(x, list) else [x]


def insert_unique(xs: Any, vs: Any) -> List[Any]:
    """
    Return copy of `xs` extended with values of `vs` that do not exist in `xs`.

    Inputs are converted to lists if they already aren't.
    """
    xs = as_array(copy.deepcopy(xs))
    vs = as_array(copy.deepcopy(vs))

    for v in vs:
        if v not in xs:
            xs.append(v)

    return cast(List[Any], xs)  # mypy doesn't recognize it


class InvalidValueType(Exception):
    def __init__(self, relative_path: str) -> None:
        if relative_path:
            super().__init__(f"The value of '{relative_path}' should be a map")
        else:
            super().__init__("It should be a map")


def dict_deep_get(d: Any, path: Union[str, List[str]]) -> Optional[Any]:
    """
    Get the value deep in the dict.

    If any value along the path doesn't exist, return None.
    If any parent node exists but is not a dict, raise InvalidValueType.
    """
    relative_path = ""
    _path_nodes = path.split(".") if isinstance(path, str) else path
    while _path_nodes:
        if d is None:
            return None
        if not isinstance(d, dict):
            raise InvalidValueType(relative_path)
        d = d.get(_path_nodes[0])
        relative_path = (relative_path + f".{_path_nodes[0]}").lstrip(".")
        _path_nodes = _path_nodes[1:]
    return d


def dict_deep_set(d: Any, path: str, value: Any) -> None:
    """
    Set the value deep in the dict.

    If any value along the path doesn't exist, set to {}.
    If any parent node exists but is not a dict, raise InvalidValueType.
    """
    relative_path = ""
    if not path:
        raise ValueError("path cannot be empty")
    _path_nodes = path.split(".")
    while len(_path_nodes) > 1:
        if not isinstance(d, dict):
            raise InvalidValueType(relative_path)
        d = d.setdefault(_path_nodes[0], {})
        relative_path = (relative_path + f".{_path_nodes[0]}").lstrip(".")
        _path_nodes = _path_nodes[1:]
    if not isinstance(d, dict):
        raise InvalidValueType(relative_path)
    d[_path_nodes[0]] = value
